import random
from datetime import datetime, timedelta
from typing import Any, Dict, Sequence, Type
from uuid import uuid4

from revjwt.jwt import encode


class PayloadDesc:
    pass


class TypDesc(PayloadDesc):
    def __init__(self, typ: str) -> None:
        self._typ = typ

    def __get__(self, instance: Any, cls: Any) -> str:
        return self._typ


class AudDesc(PayloadDesc):
    def __get__(self, instance: Any, cls: Any) -> Any:
        return instance.client.client_id


class IatDesc(PayloadDesc):
    def __get__(self, instance: Any, cls: Any) -> int:
        now = datetime.utcnow()
        return int(now.timestamp())


class JtiDesc(PayloadDesc):
    def __get__(self, instance: Any, cls: Any) -> str:
        return uuid4().__str__()


class SubDesc(PayloadDesc):
    def __get__(self, instance: Any, cls: Any) -> Any:
        return instance.data["id"]


class GrpDesc(PayloadDesc):
    def __get__(self, instance: Any, cls: Any) -> str:
        groups = instance.data.get("groups", [])
        return ":".join(groups)


class ExpDesc(PayloadDesc):
    exp_field: str = "default_access_exp"
    duration_unit: str = "minutes"
    default_exp: int = 120

    def __init__(self, exp_field: str, dur_unit: str, default_exp: int) -> None:
        self._exp_field = exp_field
        self._dur_unit = dur_unit
        self._default_exp = default_exp

    def __get__(self, instance: Any, cls: Any) -> int:
        exp = instance.client.get(self._exp_field, self._default_exp)
        now = datetime.utcnow()
        kwargs = {self._dur_unit: exp}
        duration = timedelta(**kwargs)
        real_exp = int((now + duration).timestamp())
        return real_exp


class BaseBuilder(type):
    def __new__(  # type: ignore
        cls, name: str, bases: Sequence[Any], attrs: Dict[Any, Any]
    ) -> Type["BaseBuilder"]:
        attrs["_payloads"] = [
            key for key, value in attrs.items() if isinstance(value, PayloadDesc)
        ]
        return super().__new__(cls, name, bases, attrs)  # type: ignore


class Builder(metaclass=BaseBuilder):
    _payloads: Sequence[str]

    def __init__(self, client: Any, data: Any) -> None:
        self.client = client
        self.data = data

    def get_payload(self) -> Dict[str, Any]:
        payloads = {}
        for payload in self._payloads:
            payloads[payload] = getattr(self, payload)

        return payloads


class JWTEncoder:
    payload_builder_class: Type["Builder"]
    key_class: Any

    def __init__(self, client: Any) -> None:
        self.client = client

    def get_private_key(self) -> str:
        keys = self.key_class.objects(alg="RSA256", status="enabled")
        key = random.choice(keys)
        return key["id"]  # type: ignore

    def build_payload(self, data: Any) -> Dict[str, Any]:
        builder = self.payload_builder_class(client=self.client, data=data)
        return builder.get_payload()

    def encode(self, user: Any) -> str:
        key = self.get_private_key()
        data = self.build_payload(user)
        headers = {"kid": key}
        encoded = encode(data, key=key, algorithm="RS256", headers=headers)
        return encoded



from typing import Any, Dict

from google.auth.transport import requests
from google.oauth2 import id_token


def decode(token: str, client_id: str) -> Dict[str, Any]:
    request = requests.Request()

    id_info = id_token.verify_oauth2_token(token, request, client_id)

    if id_info["iss"] != "accounts.google.com":
        raise ValueError("Wrong issuer.")

    return id_info  # type: ignore



import json
from calendar import timegm
from collections.abc import Mapping
from datetime import datetime
from typing import Any, Dict, List, Optional, Union

from jwt.api_jwt import PyJWT
from jwt.exceptions import DecodeError, InvalidAudienceError, MissingRequiredClaimError

from revjwt.jws import decode as jws_decode
from revjwt.jws import encode as jws_encode


class JWT(PyJWT):
    def encode(
        self,
        payload: Dict[str, Any],
        key: str,
        algorithm: str = "RSA256",
        headers: Optional[Dict[str, Any]] = None,
        json_encoder: Any = None,
    ) -> str:
        # Check that we get a mapping
        if not isinstance(payload, Mapping):
            raise TypeError(
                "Expecting a mapping object, as JWT only supports "
                "JSON objects as payloads."
            )

        # Payload
        payload = payload.copy()
        for time_claim in ["exp", "iat", "nbf"]:
            # Convert datetime to a intDate value in known time-format claims
            if isinstance(payload.get(time_claim), datetime):
                payload[time_claim] = timegm(payload[time_claim].utctimetuple())

        json_payload = json.dumps(
            payload, separators=(",", ":"), cls=json_encoder
        ).encode("utf-8")

        return jws_encode(json_payload, key, algorithm, headers, json_encoder)

    def decode_complete(
        self,
        jwt: str,
        key: str = "",
        algorithms: List[str] = ["RS256"],
        options: Any = None,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        algorithms = ["RS256"]
        if options is None:
            options = {"verify_signature": True}
        else:
            options.setdefault("verify_signature", True)

        if options["verify_signature"] and not algorithms:
            raise DecodeError(
                'It is required that you pass in a value for the "algorithms" argument when calling decode().'
            )

        decoded = jws_decode(
            jwt,
            key="",
            algorithms=algorithms,
            options=options,
            **kwargs,
        )

        try:
            payload = json.loads(decoded["payload"])
        except ValueError as e:
            raise DecodeError("Invalid payload string: %s" % e)
        if not isinstance(payload, dict):
            raise DecodeError("Invalid payload string: must be a json object")

        if options["verify_signature"]:
            merged_options = {**self.options, **options}
            self._validate_claims(payload, merged_options, **kwargs)  # type: ignore

        decoded["payload"] = payload
        return decoded

    def _validate_aud(
        self, payload: Dict[str, Any], audience: Union[str, List[str]]
    ) -> None:
        if audience is None:
            return
        if audience is None and "aud" not in payload:
            return

        if audience is not None and "aud" not in payload:
            # Application specified an audience, but it could not be
            # verified since the token does not contain a claim.
            raise MissingRequiredClaimError("aud")  # type: ignore

        audience_claims = payload["aud"]

        if isinstance(audience_claims, str):
            audience_claims = [audience_claims]
        if not isinstance(audience_claims, list):
            raise InvalidAudienceError("Invalid claim format in token")
        if any(not isinstance(c, str) for c in audience_claims):
            raise InvalidAudienceError("Invalid claim format in token")

        if isinstance(audience, str):
            audience = [audience]

        if not any(aud in audience_claims for aud in audience):
            raise InvalidAudienceError("Invalid audience")


_jwt = JWT()  # type: ignore
encode = _jwt.encode
decode = _jwt.decode



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['revjwt']

package_data = \
{'': ['*']}

install_requires = \
['PyJWT>=2.0.0,<3.0.0',
 'boto3>=1.17.27,<2.0.0',
 'jwcrypto>=0.8,<0.9',
 'requests>=2.25.1,<3.0.0']

setup_kwargs = {
    'name': 'revjwt',
    'version': '1.0.1',
    'description': '',
    'long_description': None,
    'author': 'Chien',
    'author_email': 'a0186163@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import json
from typing import Any, Dict, List
from urllib.parse import urlparse

import requests
from jwcrypto.jwk import JWK
from jwt.api_jws import PyJWS
from jwt.exceptions import DecodeError

from revjwt.algorithms import KMSAlgorithm

PUB_URL = "https://keys.revtel-api.com/certs.json"
PUB_STG_URL = "https://keys.revtel-api.com/certs-stg.json"
DEP_URL = "https://keys.revtel-api.com/pub.json"


class JWS(PyJWS):
    def __init__(self, options: Any = None) -> None:
        super().__init__(options)  # type: ignore
        self._algorithms = {"RS256": KMSAlgorithm()}

    def decode_complete(
        self,
        jwt: str,
        key: str = "",
        algorithms: List[str] = ["RS256"],
        options: Any = None,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        if options is None:
            options = {}
        merged_options = {**self.options, **options}
        verify_signature = merged_options["verify_signature"]

        if verify_signature and not algorithms:
            raise DecodeError(
                'It is required that you pass in a value for the "algorithms" argument when calling decode().'
            )

        payload, signing_input, header, signature = self._load(jwt)  # type: ignore

        json_payload = json.loads(payload.decode())
        iss: str = json_payload["iss"]

        if not iss.startswith("https"):
            iss = "https://" + iss

        try:
            parsed = urlparse(iss)
            host, version = parsed.netloc, parsed.path[-2:]
        except ValueError:
            host, version = DEP_URL, "v1"

        kid = header["kid"]

        if version == "v4":
            env = json_payload["env"]
        elif version == "v3":
            env = host.split(".")[0][-3:]

        if version not in ["v2", "v1"]:
            url = PUB_STG_URL if env == "stg" else PUB_URL
            resp = requests.get(url).json()["keys"]
            try:
                key = [key for key in resp if key["kid"] == kid][0]
            except IndexError:
                raise DecodeError(f"key: {kid} not found")
        else:
            resp = requests.get(DEP_URL).json()
            try:
                key = [key for key in resp if key["kid"] == kid][0]
            except IndexError:
                raise DecodeError(f"key: {kid} not found")

        key_json = JWK.from_json(json.dumps(key))
        key_pem = key_json.export_to_pem()

        self._verify_signature(signing_input, header, signature, key_pem, algorithms)  # type: ignore

        return {
            "payload": payload,
            "header": header,
            "signature": signature,
        }


_jws = JWS()
encode = _jws.encode
decode = _jws.decode_complete
